#pragma	config	OSC = HSPLL
#pragma config	WDT = ON, WDTPS = 128

/*
 *	pic18f8722 sram puf POC routines - cut from a larger project using the ET-BASE PIC18F8722 board
 *  uses hamming code routines from Michael Dipperstein (mdipper@cs.ucsb.edu) with small mods to work with  C18
 *	The chip MUST be erased and blank checked before programmed with this. After programming run the program until the LED
 *	blinks fast, then power cycle and run again at least 10 time to generate a stable bit pattern.
 *	Running the unit in a shielded box, having stable power and power switching helps reduce the chances of random bit flipping
 *	nsaspook@nsaspook.com
 */

#include <p18F8722.h>
#include <usart.h>
#include <timers.h>
#include <stdlib.h>
#include <eep.h>
#include <stdio.h>
#include <string.h>
#include "hamming.h"


void rx_handler (void);

#define BUF_SIZE 16
#define PUF_SIZE 128


#define FALSE	0
#define TRUE	1
#define	BLINK_RATE	35000


#define	DOTESTING	FALSE			// inject random data for robustness testing.
#define	TIMEROFFSET	26474			// timer0 16bit counter value for 1 second to overflow
#define	TIMERFAST	59025			// fast flash 6hz
#define	INITTIME	36
#define	COMMTIME	30

volatile int	CATCH=FALSE ,y=0, LED_UP=TRUE, z=0, TOUCH=FALSE, UNTOUCH=FALSE, LCD_OK=FALSE, TESTING=DOTESTING, comm_check=0,init_check=0,
                                      SCREEN_INIT=FALSE;
volatile long	j=0,alive_led=0;

volatile unsigned int	timer0_off=TIMEROFFSET;

//						data section for sram that contains 'random' data at power up
#pragma udata sdbank
far unsigned char	sram_key[PUF_SIZE], sram_key_masked[PUF_SIZE];
#pragma udata

#pragma idata bigdata
char  	testing_data,key_string[16];


const rom char		*build_date = __DATE__, *build_time = __TIME__;

/******************************************************************************
*  Make CRC 16 Look-up Table.                                            *
******************************************************************************/

const unsigned int far crc_table[0x100] = {
        0x0000, 0x1021, 0x2042, 0x3063, 0x4084, 0x50A5, 0x60C6, 0x70E7,
        0x8108, 0x9129, 0xA14A, 0xB16B, 0xC18C, 0xD1AD, 0xE1CE, 0xF1EF,
        0x1231, 0x0210, 0x3273, 0x2252, 0x52B5, 0x4294, 0x72F7, 0x62D6,
        0x9339, 0x8318, 0xB37B, 0xA35A, 0xD3BD, 0xC39C, 0xF3FF, 0xE3DE,
        0x2462, 0x3443, 0x0420, 0x1401, 0x64E6, 0x74C7, 0x44A4, 0x5485,
        0xA56A, 0xB54B, 0x8528, 0x9509, 0xE5EE, 0xF5CF, 0xC5AC, 0xD58D,
        0x3653, 0x2672, 0x1611, 0x0630, 0x76D7, 0x66F6, 0x5695, 0x46B4,
        0xB75B, 0xA77A, 0x9719, 0x8738, 0xF7DF, 0xE7FE, 0xD79D, 0xC7BC,
        0x48C4, 0x58E5, 0x6886, 0x78A7, 0x0840, 0x1861, 0x2802, 0x3823,
        0xC9CC, 0xD9ED, 0xE98E, 0xF9AF, 0x8948, 0x9969, 0xA90A, 0xB92B,
        0x5AF5, 0x4AD4, 0x7AB7, 0x6A96, 0x1A71, 0x0A50, 0x3A33, 0x2A12,
        0xDBFD, 0xCBDC, 0xFBBF, 0xEB9E, 0x9B79, 0x8B58, 0xBB3B, 0xAB1A,
        0x6CA6, 0x7C87, 0x4CE4, 0x5CC5, 0x2C22, 0x3C03, 0x0C60, 0x1C41,
        0xEDAE, 0xFD8F, 0xCDEC, 0xDDCD, 0xAD2A, 0xBD0B, 0x8D68, 0x9D49,
        0x7E97, 0x6EB6, 0x5ED5, 0x4EF4, 0x3E13, 0x2E32, 0x1E51, 0x0E70,
        0xFF9F, 0xEFBE, 0xDFDD, 0xCFFC, 0xBF1B, 0xAF3A, 0x9F59, 0x8F78,
        0x9188, 0x81A9, 0xB1CA, 0xA1EB, 0xD10C, 0xC12D, 0xF14E, 0xE16F,
        0x1080, 0x00A1, 0x30C2, 0x20E3, 0x5004, 0x4025, 0x7046, 0x6067,
        0x83B9, 0x9398, 0xA3FB, 0xB3DA, 0xC33D, 0xD31C, 0xE37F, 0xF35E,
        0x02B1, 0x1290, 0x22F3, 0x32D2, 0x4235, 0x5214, 0x6277, 0x7256,
        0xB5EA, 0xA5CB, 0x95A8, 0x8589, 0xF56E, 0xE54F, 0xD52C, 0xC50D,
        0x34E2, 0x24C3, 0x14A0, 0x0481, 0x7466, 0x6447, 0x5424, 0x4405,
        0xA7DB, 0xB7FA, 0x8799, 0x97B8, 0xE75F, 0xF77E, 0xC71D, 0xD73C,
        0x26D3, 0x36F2, 0x0691, 0x16B0, 0x6657, 0x7676, 0x4615, 0x5634,
        0xD94C, 0xC96D, 0xF90E, 0xE92F, 0x99C8, 0x89E9, 0xB98A, 0xA9AB,
        0x5844, 0x4865, 0x7806, 0x6827, 0x18C0, 0x08E1, 0x3882, 0x28A3,
        0xCB7D, 0xDB5C, 0xEB3F, 0xFB1E, 0x8BF9, 0x9BD8, 0xABBB, 0xBB9A,
        0x4A75, 0x5A54, 0x6A37, 0x7A16, 0x0AF1, 0x1AD0, 0x2AB3, 0x3A92,
        0xFD2E, 0xED0F, 0xDD6C, 0xCD4D, 0xBDAA, 0xAD8B, 0x9DE8, 0x8DC9,
        0x7C26, 0x6C07, 0x5C64, 0x4C45, 0x3CA2, 0x2C83, 0x1CE0, 0x0CC1,
        0xEF1F, 0xFF3E, 0xCF5D, 0xDF7C, 0xAF9B, 0xBFBA, 0x8FD9, 0x9FF8,
        0x6E17, 0x7E36, 0x4E55, 0x5E74, 0x2E93, 0x3EB2, 0x0ED1, 0x1EF0
};

volatile unsigned int tchar, uchar;
#pragma idata

#pragma code rx_interrupt = 0x8
void rx_int (void)
{
        _asm goto rx_handler _endasm
}
#pragma code

#pragma interrupt rx_handler
void rx_handler (void)							// all serial data transform functions are handled here
{
        static unsigned char c1=0,c2=0,testc=0;
        static int i=0;

        ClrWdt();										// reset the WDT timer

        if ( INTCONbits.TMR0IF ) {								// check timer0 irq time timer
                //check for TMR0 overflow
                INTCONbits.TMR0IF = 0; 							//clear interrupt flag
                WriteTimer0 ( timer0_off );
                PORTHbits.RH0=!PORTHbits.RH0;					// flash onboard led
                testc= (unsigned char) rand();
        }

        if (PIR1bits.RCIF) {							// is data from host COMM1
                if ( RCSTA1bits.OERR ) {
                        RCSTA1bits.CREN = 0;			//	clear overrun
                        RCSTA1bits.CREN = 1;			// re-enable
                }

                PIR1bits.RC1IF = 0;						// clear irq flag
        }

        if (PIR3bits.RC2IF || TESTING) {							// is data from  COMM2
                if ( RCSTA2bits.OERR ) {
                        RCSTA2bits.CREN = 0;			//	clear overrun
                        RCSTA2bits.CREN = 1;			// re-enable
                }

                /* Clear the interrupt flag */
                PIR3bits.RC2IF = 0;						// clear usart 2 flag
        }
}

void wdtdelay(unsigned long delay)
{
        unsigned long dcount;
        for (dcount=0; dcount<=delay; dcount++) {			// delay a bit
                ClrWdt();									// reset the WDT timer
        };
}


/*	Count bits 	*/

unsigned char	bit_count(unsigned char input)
{
        unsigned char counter=0,i;

        for	(i=0; i<8; i++) {
                if ( (input & (1<<i)) != 0 ) counter++;
        }
        return (counter);
}

/******************************************************************************
*  Make CRC 16 Calculation Function.                                     *
******************************************************************************/

unsigned int Make_Crc16 (unsigned char *data, unsigned int length)
{
        unsigned int accum,len;

        /* Pre-conditioning */
        accum = 0xFFFF;
        len=length;


        ClrWdt();									// reset the WDT timer
        while (len--) {
                accum = ((accum & 0x00FF) << 8) ^
                        crc_table[( (accum >> 8) ^ *data ++ ) & 0x00FF];
        }

        ClrWdt();									// reset the WDT timer
        return(accum);
}


/* get random data for sram powerup bits */
unsigned int puf_sram(unsigned char cmode, unsigned char kmode)						// look at random SRAM data for PRNG seeds and ID key
{																// uses a udata section of memeory 2xPUF_SIZE and eeprom from
        unsigned int	e;										// 0 to 3xPUF_SIZE to store key, diff data, masked data
        static unsigned int	seed=0, seeda=0, seedk=0;			// if cmode is TRUE the stored key diff data will be zeroed and will return 0
        unsigned char	entr_s,entr_r,entr_d;					// if kmode is TRUE the seed will be generated from the masked contents of sram
        // if kmode is FALSE the seed will be generated from the raw sram contents
        // kmode should be stable after about 10 or more power cycles

        for (e=0; e<PUF_SIZE; e++) {
                seeda = seeda + sram_key[e];
                entr_s=sram_key[e];
                Busy_eep();
                entr_r = Read_b_eep ( e );					// read eeprom seed data
                Busy_eep();
                entr_d = Read_b_eep ( e+PUF_SIZE );			// read unstuck bits data
                entr_d = entr_d | (entr_s^entr_r);			// XOR to look for diff bits
                sram_key_masked[e]=sram_key[e] & (~entr_d);	// store unstuck bits

                if (!cmode) {								// cmode  will zero key data history
                        Busy_eep();
                        Write_b_eep ( e, entr_s );				// write eeprom new key data
                        Busy_eep();
                        Write_b_eep ( e+PUF_SIZE, entr_d );		// write eeprom key unstuck bits data
                        Busy_eep();
                        Write_b_eep ( e+PUF_SIZE+PUF_SIZE, sram_key_masked[e] );		// write eeprom key stuck bits data
                } else {
                        Busy_eep();
                        Write_b_eep ( e, entr_s );				// write eeprom key data
                        Busy_eep();
                        Write_b_eep ( e+PUF_SIZE, 0 );			// write zeros to  eeprom key unstuck bits data
                        Busy_eep();
                        Write_b_eep ( e+PUF_SIZE+PUF_SIZE, 0 );	// write zeros to  eeprom key masked data
                }
                ClrWdt();									// reset the WDT timer
        }

        seed=Make_Crc16(sram_key,PUF_SIZE);										// make seed from crc16 of sram
        if (kmode) seed=Make_Crc16(sram_key_masked,PUF_SIZE);					// make seed from crc16 of sram masked with unstuck bits
        if (cmode) seed=0; 														// cmode return 0
        return (seed);
}

void main (void)
{
        unsigned char	cylon=0xfe,blanked,hammed;
        unsigned int	pic_key=PUF_SIZE*3,i;
        /* Configure  PORT pins for output */
        TRISB = 0;
        TRISJ = 0;
        TRISH = 0;
        PORTHbits.RH0=1;					// onboard led OFF
        TRISE = 0;
        TRISD = 0;
        PORTE = 0xFF;
        TRISF = 0;
        PORTF = 0xFF;
        sram_key[0]=0x0;						// set PUF SRAM address 0 byte to 0

        /*
         * Open the USART configured as
         * 8N1, 9600 baud, in receive INT mode
         */
        Open1USART (USART_TX_INT_OFF &
                    USART_RX_INT_OFF &			//	no receive int
                    USART_ASYNCH_MODE &
                    USART_EIGHT_BIT &
                    USART_CONT_RX &
                    USART_BRGH_LOW, 64);				// 40mhz osc HS		9600 baud

        Open2USART (USART_TX_INT_OFF &
                    USART_RX_INT_OFF &			// no receive int
                    USART_ASYNCH_MODE &
                    USART_EIGHT_BIT &
                    USART_CONT_RX &
                    USART_BRGH_LOW, 64);				// 40mhz osc HS		9600 baud

        OpenTimer0 ( TIMER_INT_ON & T0_16BIT & T0_SOURCE_INT & T0_PS_1_256 );
        WriteTimer0 ( timer0_off );						//	start timer0 at 1 second ticks

        /* Enable interrupt priority */
        RCONbits.IPEN = 1;

        /* Make receive interrupt high priority */
        IPR1bits.RC1IP = 1;
        IPR3bits.RC2IP = 1;

        while (DataRdy1USART()) {		// dump rx data
                z = Read1USART();
        };
        while (DataRdy2USART()) {		// dump rx data
                z = Read2USART();
        };

        PORTD=0x00;
        alive_led=0xfe;
        Busy_eep();
        blanked = Read_b_eep ( 0 );						// read eeprom seed data 0
        if (blanked != 0) {
                puf_sram(TRUE,FALSE);			// set puf eeprom areas to 0's
        } else {
                srand(puf_sram(FALSE,FALSE));					// use unmasked seed for prng
        }

        PORTF = ~HammingTableDecode(sram_key_masked[0x0f] & 0x7F);							// debug display, only use 7 bits

        /*		Store the hamming code corrected PIC chip key in the PUF_SIZE*3 eeprom memory
         *		after a few power cycles of the chip, this key code should be stable and usable to
         *		identify it from a random chip of the same type (YMMV).
         */
        for (i=0; i<PUF_SIZE; i+=2) {
                hammed=HammingTableDecode(sram_key_masked[i] & 0x7F);							// only use 7 bits
                hammed= (hammed<<4) + HammingTableDecode(sram_key_masked[i+1] & 0x7F);			// only use 7 bits
                Busy_eep();
                Write_b_eep ( pic_key, hammed );	// write PIC chip PUF key to eeprom
                pic_key++;
                sprintf(key_string,"%d,",hammed);
                puts1USART(key_string);				// send the codes via rs-232 port
        }
        sprintf(key_string,"\r\n");
        puts1USART(key_string);				// terminate

        /* Enable all high priority interrupts */
        INTCONbits.GIEH = 1;
        INTCONbits.TMR0IE=1;

        /* Loop forever */

        while (TRUE) {							// busy work

                if (j++ >= BLINK_RATE) {								// delay a bit ok

                        INTCONbits.GIEH = 0;
                        if (LCD_OK) {									// screen status feedback
                                LATD = ~cylon;								// roll leds cylon style
                                timer0_off=TIMEROFFSET;
                                TESTING=FALSE;
                        } else {
                                LATD = cylon;								// roll leds cylon style (inverted)
                                timer0_off=TIMERFAST;
                        }
                        INTCONbits.GIEH = 1;

                        if (LED_UP && (alive_led!=0)) {
                                alive_led=alive_led*2;
                                cylon = cylon <<1;
                        } else {
                                if (alive_led!=0) alive_led=alive_led/2;
                                cylon = cylon >>1;
                        }
                        if (alive_led<2) {
                                alive_led=2;
                                LED_UP=TRUE;
                        } else {
                                if (alive_led>128) {
                                        alive_led=128;
                                        LED_UP=FALSE;
                                }
                        }
                        j=0;
                        ClrWdt();								// reset the WDT timer
                }
        }
        ;
}
